@@ -2,7 +2,7 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-from keras.layers import InputLayer
+from tensorflow.python.keras.layers import InputLayer
 
 from ..utils import to_list, floatx, set_floatx
 from .rnn_functional import RNNFunctional
@@ -51,3 +51,7 @@ class RNNVariable(RNNFunctional):
             inputs=to_list(layer.input),
             outputs=to_list(layer.output),
         )
+
+    @classmethod
+    def get_class(cls):
+        return RNNFunctional
\ No newline at end of file
