@@ -8,21 +8,23 @@ from __future__ import print_function
 import os
 
 import tensorflow as tf
-import tensorflow.python.keras.backend as K
-import tensorflow.python.keras as k
+from tensorflow.python.keras import backend as K
+from tensorflow.python import keras as k
 import numpy as np
 
+from tensorflow.python.keras.utils.data_utils import Sequence
 from tensorflow.python.keras.models import Model
 from tensorflow.python.keras.utils.vis_utils import plot_model
 from tensorflow import gradients as tf_gradients
 
 from ..utils import unpack_singleton, to_list
 from ..utils import is_variable, is_constraint, is_functional
-from ..utils.optimizers import GradientObserver, ScipyOptimizer
+from ..utils.optimizers import GradientObserver, ScipyOptimizer, GeneratorWrapper
+from ..utils.callbacks import *
 
 from ..functionals import Variable
 from ..functionals import RadialBasis
-from ..constraints import Data, Tie
+from ..constraints import Data, PDE, Tie
 
 
 class SciModel(object):
@@ -96,6 +98,7 @@ class SciModel(object):
             if not is_constraint(y):
                 if is_functional(y):
                     # Case of Data-type constraint.
+                    # By default, targets are initialized with Data.
                     targets[i] = Data(y)
                 elif isinstance(y, tuple) and \
                         len(y) == 2 and \
@@ -126,34 +129,29 @@ class SciModel(object):
             outputs=output_vars,
             **kwargs
         )
+
         # compile the model.
         loss_weights = [K.variable(1.0) for v in output_vars]
-        if isinstance(optimizer, str) and \
-                len(optimizer.lower().split("scipy-")) > 1:
-            model.compile(
-                loss=loss_func,
-                optimizer=GradientObserver(method=optimizer),
-                loss_weights=loss_weights
-            )
-        else:
-            model.compile(
-                loss=loss_func,
-                optimizer=optimizer,
-                loss_weights=loss_weights
-            )
-        # model.train_function = True
 
+        # Set the variables.
+        self._model = model
+        self._inputs = inputs
+        self._constraints = targets
+        self._loss_func = loss_func
+        self._optimizer = optimizer
+        self._loss_weights = loss_weights
+        self._callbacks = {}
+        
+        # compile the model 
+        self.compile()
+        
         # set initial state of the model.
         if load_weights_from is not None:
             if os.path.exists(load_weights_from): 
                 model.load_weights(load_weights_from)
             else:
                 raise Warning("File not found - load_weights_from: {}".format(load_weights_from))
-        # Set the variables.
-        self._model = model
-        self._inputs = inputs
-        self._constraints = targets
-        self._loss_func = loss_func
+
         # Plot to file if requested.
         if plot_to_file is not None:
             plot_model(self._model, to_file=plot_to_file)
@@ -170,6 +168,22 @@ class SciModel(object):
     def inputs(self):
         return self._inputs
 
+    def compile(self):
+        loss_func = self._loss_func 
+        loss_weights = self._loss_weights
+        optimizer = self._optimizer
+
+        if isinstance(optimizer, str) and \
+                len(optimizer.lower().split("scipy-")) > 1:
+            optimizer = GradientObserver(method=optimizer)
+
+        self._model.compile(loss=loss_func,
+                            optimizer=optimizer,
+                            loss_weights=loss_weights)
+
+    def clear_callbacks(self):
+        self._callbacks.clear()
+
     def load_weights(self, file):
         if os.path.exists(file):
             self.model.load_weights(file)
@@ -209,13 +223,15 @@ class SciModel(object):
 
     def train(self,
               x_true,
-              y_true,
+              y_true=None,
               weights=None,
               target_weights=None,
               batch_size=2**6,
               epochs=100,
               learning_rate=0.001,
-              adaptive_weights=False,
+              adaptive_weights=None,
+              adaptive_sample_weights=None,
+              log_loss_gradients=None,
               shuffle=True,
               callbacks=None,
               stop_lr_value=1e-8,
@@ -223,6 +239,9 @@ class SciModel(object):
               reduce_lr_min_delta=0.,
               stop_after=None,
               stop_loss_value=1e-8,
+              log_parameters=None,
+              log_functionals=None,
+              log_loss_landscape=None,
               save_weights_to=None,
               save_weights_freq=0,
               default_zero_weight=0.0,
@@ -232,6 +251,7 @@ class SciModel(object):
 
         # Arguments
             x_true: list of `Xs` associated to targets of `Y`.
+                Alternatively, you can pass a Sequence object (keras.utils.Sequence).
                 Expecting a list of np.ndarray of size (N,1) each,
                 with N as the sample size.
             y_true: list of true `Ys` associated to the targets defined during model setup.
@@ -256,8 +276,14 @@ class SciModel(object):
                     learning_rate = ([0, 100, 1000], [0.001, 0.0005, 0.00001])
             shuffle: Boolean (whether to shuffle the training data).
                 Default value is True.
-            adaptive_weights: Defaulted to False (no updates - evaluated once in the beginning).
-                Used if the model is compiled with adaptive_weights.
+            adaptive_weights: Pass a Dict with the following keys:
+                . method: GP or NTK.
+                . freq: Freq to update the weights.
+                . log_freq: Freq to log the weights and gradients in the history object.
+                . beta: The beta parameter in from Gradient Pathology paper.
+            log_loss_gradients: Pass a Dict with the following keys:
+                . freq: Freq of logs. Defaulted to 100.
+                . path: Path to log the gradients.
             callbacks: List of `keras.callbacks.Callback` instances.
             reduce_lr_after: patience to reduce learning rate or stop after certain missed epochs.
                 Defaulted to epochs max(10, epochs/10).
@@ -268,7 +294,21 @@ class SciModel(object):
                 This values affects number of failed attempts to trigger reduce learning rate based on reduce_lr_after. 
             stop_after: To stop after certain missed epochs. Defaulted to total number of epochs.
             stop_loss_value: The minimum value of the total loss that stops the training automatically. 
-                Defaulted to 1e-8. 
+                Defaulted to 1e-8.
+            log_parameters: Dict object expecting the following keys:
+                . parameters: pass list of parameters.
+                . freq: pass freq of outputs.
+            log_functionals: Dict object expecting the following keys:
+                . functionals: List of functionals to log their training history.
+                . inputs: The input grid to evaluate the value of each functional.
+                          Should be of the same size as the inputs to the model.train.
+                . path: Path to the location that the csv files will be logged.
+                . freq: Freq of logging the functionals.
+            log_loss_landscape: Dict object expecting the following arguments:
+                . norm: defaulted to 2.
+                . resolution: defaulted to 10.
+                . path: Path to the location that the csv files will be logged.
+                . freq: Freq of logging the loss landscape.
             save_weights_to: (file_path) If you want to save the state of the model (at the end of the training).
             save_weights_freq: (Integer) Save weights every N epcohs.
                 Defaulted to 0.
@@ -277,6 +317,73 @@ class SciModel(object):
         # Returns
             A Keras 'History' object after performing fitting.
         """
+        if isinstance(x_true, Sequence):
+            if y_true is not None:
+                raise ValueError(
+                    'If data_generator is provided, you should not provide y_true.'
+                )
+            data_generator = x_true
+
+        else:
+            if y_true is None:
+                raise ValueError('(X_true, Y_true): Please provide proper values for both inputs and targets of SciModel. ')
+            # prepare X,Y data.
+            x_true = to_list(x_true)
+            for i, (x, xt) in enumerate(zip(x_true, self._model.inputs)):
+                x_shape = tuple(xt.get_shape().as_list())
+                if x.shape != x_shape:
+                    try:
+                        x_true[i] = x.reshape((-1,) + x_shape[1:])
+                    except:
+                        print(
+                            'Could not automatically convert the inputs to be ' 
+                            'of the same size as the expected input tensors. ' 
+                            'Please provide inputs of the same dimension as the `Variables`. '
+                        )
+                        assert False
+
+            y_true = to_list(y_true)
+            assert len(y_true)==len(self._constraints), \
+                'Miss-match between expected targets (constraints) defined in `SciModel` and ' \
+                'the provided `y_true`s - expecting the same number of data points. '
+
+            num_sample = x_true[0].shape[0]
+            assert all([x.shape[0]==num_sample for x in x_true[1:]]), \
+                'Inconsistent sample size among `Xs`. '
+            ids_all = np.arange(0, num_sample)
+
+            if weights is None:
+                weights = np.ones(num_sample)
+            elif isinstance(weights, np.ndarray):
+                if len(weights.shape)!=1 or \
+                        weights.shape[0] != num_sample:
+                    try:
+                        weights = weights.reshape(num_sample)
+                    except:
+                        raise ValueError(
+                            'Input error: `weights` should have dimension 1 with '
+                            'the same sample length as `Xs. '
+                        )
+
+            sample_weights, y_star = [], []
+            for i, yt in enumerate(y_true):
+                c = self._constraints[i]
+                # verify entry.
+                ys, wei = SciModel._prepare_data(
+                    c.cond.outputs, to_list(yt),
+                    weights if isinstance(weights, np.ndarray) else weights[i],
+                    num_sample, default_zero_weight
+                )
+                # add to the list.
+                y_star += ys
+                sample_weights += wei
+
+            # create Sequence wrapper for training. 
+            data_generator = GeneratorWrapper(
+                x_true, y_star, sample_weights, 
+                batch_size, shuffle
+            )
+
         if callbacks is None:
             # default_lr = learning_rate/0.9900
             # f0 = np.log(1.0/0.9900 - 1.0)
@@ -292,15 +399,20 @@ class SciModel(object):
             if isinstance(learning_rate, (type(None), float, int)):
                 lr_rates = 0.001 if learning_rate is None else learning_rate
                 K.set_value(self.model.optimizer.lr, lr_rates)
-                callbacks.append(
-                    k.callbacks.ReduceLROnPlateau(
-                        monitor='loss', factor=0.5,
-                        patience=reduce_lr_after, #cooldown=epochs/10,
-                        verbose=1, mode='auto', 
-                        min_delta=reduce_lr_min_delta,
-                        min_lr=0.
+                if 'lr_scheduler' in self._callbacks and learning_rate is None:
+                    callbacks.append(self._callbacks['lr_scheduler'])
+                else:
+                    callbacks.append(
+                        k.callbacks.ReduceLROnPlateau(
+                            monitor='loss', factor=0.5,
+                            patience=reduce_lr_after, #cooldown=epochs/10,
+                            verbose=1, mode='auto',
+                            min_delta=reduce_lr_min_delta,
+                            min_lr=0.
+                        )
                     )
-                )
+                    self._callbacks['lr_scheduler'] = callbacks[-1]
+
             elif isinstance(learning_rate, (tuple, list)):
                 lr_epochs = learning_rate[0]
                 lr_rates = learning_rate[1]
@@ -308,80 +420,34 @@ class SciModel(object):
                     # k.callbacks.LearningRateScheduler(lambda n: default_lr/(1.0 + np.exp((n-n0)/a0))),
                     k.callbacks.LearningRateScheduler(lambda n: np.interp(n, lr_epochs, lr_rates))
                 )
+                self._callbacks['lr_scheduler'] = callbacks[-1]
+
             else:
                 raise ValueError(
                     "learning rate: expecting a `float` or a tuple/list of two arrays"
                     " with `epochs` and `learning rates`"
                 )
+
             callbacks += [
                 k.callbacks.EarlyStopping(monitor="loss", mode='auto', verbose=1,
                                           patience=stop_after, min_delta=1e-9),
                 k.callbacks.TerminateOnNaN(),
                 EarlyStoppingByLossVal(stop_loss_value),
                 EarlyStoppingByLearningRate(stop_lr_value),
+                EpochTime()
             ]
-        # prepare X,Y data.
-        x_true = to_list(x_true)
-        for i, (x, xt) in enumerate(zip(x_true, self._model.inputs)):
-            x_shape = tuple(xt.get_shape().as_list())
-            if x.shape != x_shape:
-                try:
-                    x_true[i] = x.reshape((-1,) + x_shape[1:])
-                except:
-                    print(
-                        'Could not automatically convert the inputs to be ' 
-                        'of the same size as the expected input tensors. ' 
-                        'Please provide inputs of the same dimension as the `Variables`. '
-                    )
-                    assert False
-
-        num_sample = x_true[0].shape[0]
-        assert all([x.shape[0]==num_sample for x in x_true[1:]]), \
-            'Inconsistent sample size among `Xs`. '
-        ids_all = np.arange(0, num_sample)
-
-        if weights is None:
-            weights = np.ones(num_sample)
-        else:
-            if len(weights.shape)!=1 or \
-                    weights.shape[0] != num_sample:
-                try:
-                    weights = weights.reshape(num_sample)
-                except:
-                    raise ValueError(
-                        'Input error: `weights` should have dimension 1 with '
-                        'the same sample length as `Xs. '
-                    )
-
-        y_true = to_list(y_true)
-        assert len(y_true)==len(self._constraints), \
-            'Miss-match between expected targets (constraints) defined in `SciModel` and ' \
-            'the provided `y_true`s - expecting the same number of data points. '
-
-        sample_weights, y_star = [], []
-        for i, yt in enumerate(y_true):
-            c = self._constraints[i]
-            # verify entry.
-            ys, wei = SciModel._prepare_data(
-                c.cond.outputs, to_list(yt),
-                weights, num_sample, default_zero_weight
-            )
-            # add to the list.
-            y_star += ys
-            sample_weights += wei
-
+        
+        len_inputs = len(to_list(self._model.inputs))
+        len_outputs = len(to_list(self._model.outputs))
         if target_weights is not None:
             if not(isinstance(target_weights, list) and
-                   len(target_weights) == len(y_true)):
+                   len(target_weights) == len_outputs):
                 raise ValueError(
                     'Expected a list of weights for the same size as the targets '
                     '- was provided {}'.format(target_weights)
                 )
-            else:
-                for i, cw in enumerate(target_weights):
-                    sample_weights[i] *= cw
         else:
-            target_weights = len(y_true) * [1.0]
+            target_weights = len_outputs * [1.0]
 
         # save model.
         model_file_path = None
@@ -406,21 +472,113 @@ class SciModel(object):
             opt_fit_func = self._model.fit
 
         if adaptive_weights:
+            if not isinstance(adaptive_weights, dict):
+                adaptive_weights = NTKLossWeight.prepare_inputs(adaptive_weights)
+            if adaptive_weights["method"].upper() == "GP":
+                callbacks.append(
+                    GradientPathologyLossWeight(
+                        self.model, data_generator=data_generator, 
+                        types=[type(v).__name__ for v in self.constraints],
+                        **adaptive_weights
+                    ),
+                )
+            elif adaptive_weights["method"].upper() == "NTK":
+                callbacks.append(
+                    NTKLossWeight(
+                        self.model, data_generator=data_generator, 
+                        types=[type(v).__name__ for v in self.constraints],
+                        **adaptive_weights
+                    ),
+                )
+            elif adaptive_weights["method"].upper() == "MTL":
+                callbacks.append(
+                    MTLLossWeight(
+                        self.model, data_generator=data_generator,
+                        **adaptive_weights
+                    ),
+                )
+            elif adaptive_weights["method"].upper() == "SCORE":
+                callbacks.append(
+                    ScoreLossWeight(
+                        self.model, data_generator=data_generator,
+                        **adaptive_weights
+                    ),
+                )
+            else:
+                raise ValueError(
+                    'adaptive_weights method should be either of the followings: (GP, NTK)'
+                )
+            self._callbacks['adaptive_weights'] = callbacks[-1]
+
+        elif 'adaptive_weights' in self._callbacks:
+            callbacks.append(self._callbacks['adaptive_weights'])
+
+        if adaptive_sample_weights:
+            # sample_weights = [K.variable(wi) for wi in sample_weights]
             callbacks.append(
-                GradientPathologyLossWeight(
-                    self._model, x_true, y_star, sample_weights,
-                    beta=0.8, freq=adaptive_weights
-                ),
-                # k.callbacks.CSVLogger('GP.log')
+                AdaptiveSampleWeight2(
+                    self.model, data_generator=data_generator,
+                    types=[type(v).__name__ for v in self.constraints],
+                    **adaptive_sample_weights
+                )
             )
+            # loss_gradients = NTKSW.eval_diag_ntk()
+            # sample_weights = NTKSW.eval_sample_weights(loss_gradients)
+            self._callbacks['adaptive_sample_weights'] = callbacks[-1]
+
+        elif 'adaptive_sample_weights' in self._callbacks:
+            callbacks.append(self._callbacks['adaptive_sample_weights'])
+
+        if log_loss_gradients:
+            if not isinstance(log_loss_gradients, dict):
+                log_loss_gradients = LossGradientHistory.prepare_inputs(log_loss_gradients)
+            callbacks.append(
+                LossGradientHistory(
+                    self.model, data_generator=data_generator,
+                    **log_loss_gradients
+                )
+            )
+            self._callbacks['log_loss_gradients'] = callbacks[-1]
+        elif 'log_loss_gradients' in self._callbacks:
+            callbacks.append(self._callbacks['log_loss_gradients'])
+
+        if log_parameters:
+            if not isinstance(log_parameters, dict):
+                log_parameters = ParameterHistory.prepare_inputs(log_parameters)
+            callbacks.append(
+                ParameterHistory(**log_parameters)
+            )
+            self._callbacks['log_parameters'] = callbacks[-1]
+        elif 'log_parameters' in self._callbacks:
+            callbacks.append(self._callbacks['log_parameters'])
+
+        if log_functionals:
+            if not isinstance(log_functionals, dict):
+                log_functionals = FunctionalHistory.prepare_inputs(log_functionals)
+            callbacks.append(
+                FunctionalHistory(**log_functionals)
+            )
+            self._callbacks['log_functionals'] = callbacks[-1]
+        elif 'log_functionals' in self._callbacks:
+            callbacks.append(self._callbacks['log_functionals'])
+
+        if log_loss_landscape:
+            if not isinstance(log_loss_landscape, dict):
+                log_loss_landscape = LossLandscapeHistory.prepare_inputs(log_loss_landscape)
+            callbacks.append(
+                LossLandscapeHistory(
+                    self.model, data_generator=data_generator,
+                    **log_loss_landscape
+                )
+            )
+            self._callbacks['log_loss_landscape'] = callbacks[-1]
+        elif 'log_loss_landscape' in self._callbacks:
+            callbacks.append(self._callbacks['log_loss_landscape'])
 
         # training the models.
         history = opt_fit_func(
-            x_true, y_star,
-            sample_weight=sample_weights,  # sums to number of samples.
+            data_generator,  # sums to number of samples.
             epochs=epochs,
-            batch_size=batch_size,
-            shuffle=shuffle,
             callbacks=callbacks,
             validation_data=validation_data,
             **kwargs
@@ -465,7 +623,8 @@ class SciModel(object):
                 "Expected {} - provided {}".format(len(self._inputs), len(to_list(xs)))
             )
         # To have unified output for postprocessing - limitted support.
-        shape_default = xs[0].shape if all([x.shape==xs[0].shape for x in xs]) else None
+        shape_default = [x.shape for x in xs]
+        assert all([shape_default[0][0]==x[0] for x in shape_default[1:]])
         # prepare X,Y data.
         for i, (x, xt) in enumerate(zip(xs, self._model.inputs)):
             x_shape = tuple(xt.get_shape().as_list())
@@ -480,14 +639,20 @@ class SciModel(object):
                     )
                     assert False
 
-        y_pred = self._model.predict(xs, batch_size, verbose, steps)
+        y_pred = to_list(self._model.predict(xs, batch_size, verbose, steps))
 
-        if shape_default is not None:
+        # return uniform shapes. 
+        if all([shape_default[0]==sd for sd in shape_default[1:]]):
             try:
-                y_pred = [y.reshape(shape_default) for y in y_pred]
+                y_pred = [y.reshape(shape_default[0]) for y in y_pred]
             except:
                 print("Input and output dimensions need re-adjustment for post-processing.")
 
+        # revert back to normal.
+        for i, sd in enumerate(shape_default):
+            xs[i] = xs[i].reshape(sd)
+        xs = unpack_singleton(xs)
+
         return unpack_singleton(y_pred)
 
     def eval(self, *args):
@@ -507,10 +672,10 @@ class SciModel(object):
             if not isinstance(var_name, str):
                 raise ValueError("Value Error: Expected a LayerName as the input. ")
             x_data = to_list(args[1])
-            new_model = Model(self.inputs, self.get_layer(var_name).output)
+            new_model = K.function(self._model.inputs, self._model.get_layer(var_name).output)
             if not all([isinstance(xi, np.ndarray) for xi in x_data]):
                 raise ValueError("Please provide input data to the network. ")
-            return unpack_singleton(new_model.predict(x_data))
+            return unpack_singleton(new_model(x_data))
 
     def plot_model(self, *args, **kwargs):
         """ Keras plot_model functionality.
@@ -624,99 +789,3 @@ class SciModel(object):
 
         return ys, weis
 
-
-class EarlyStoppingByLossVal(k.callbacks.Callback):
-    def __init__(self, value, stop_after=1):
-        super(k.callbacks.Callback, self).__init__()
-        self.value = value
-        self.wait = stop_after
-        
-    def on_epoch_end(self, epoch, logs={}):
-        current = logs.get('loss')
-        if current < self.value:
-            self.wait -= 1
-            if self.wait <= 0:
-                self.model.stop_training = True
-                print("Epoch {:05d}: early stopping at loss value {:0.6e}".format(epoch, current))
-                print("Revise 'stop_loss_value={:0.12f}' in '.train' if it was not your intent. ".format(self.value))
-
-
-class EarlyStoppingByLearningRate(k.callbacks.Callback):
-    def __init__(self, value):
-        super(k.callbacks.Callback, self).__init__()
-        self.value = value
-
-    def on_epoch_end(self, epoch, logs={}):
-        current = logs.get('lr')
-        if current < self.value:
-            self.model.stop_training = True
-            print("Epoch {:05d}: early stopping at learning rate {:0.6e}".format(epoch, current))
-            print("Revise 'stop_lr_value={:0.12f}' in '.train' if it was not your intent. ".format(self.value))
-
-
-class GradientPathologyLossWeight(k.callbacks.Callback):
-    def __init__(self, model, inputs, targets, weights, beta=0.8, freq=100):
-        super(k.callbacks.Callback, self).__init__()
-        self.inputs = inputs
-        self.targets = targets
-        self.weights = weights
-        self.loss_grads = []
-        for i in range(len(model.outputs)):
-            yp = model.outputs[i]
-            ys = targets[i]
-            ws = weights[i]
-            f = tf.reduce_mean(
-                model.loss_functions[i](ys, yp, sample_weight=ws)
-            )
-            gf = tf_gradients(f, model.trainable_weights, unconnected_gradients='zero')
-            self.loss_grads.append(
-                K.function(model.inputs, gf)
-            )
-        self.freq = 0 if isinstance(freq, bool) else freq
-        self.beta = beta
-
-    def on_train_begin(self, logs={}):
-        # update loss-weights
-        self.update_loss_weights(0)
-
-    def on_epoch_end(self, epoch, logs={}):
-        if epoch > 1 and self.freq > 1 \
-                and epoch % self.freq == 0:
-            self.update_loss_weights(epoch)
-        # log the weights in the history output.
-        logs['adaptive_weights'] = [
-            K.get_value(wi) for wi in self.model.loss_weights
-        ]
-
-    def eval_loss_weights(self):
-        # eval new gradients
-        updated_grads = []
-        for lgi, trg in zip(self.loss_grads, self.targets):
-            updated_grads.append(
-                np.concatenate(
-                    [np.abs(wg).flatten() for wg in lgi(self.inputs)]
-                )
-            )
-        # eval max normalization on PDE.
-        ref_grad = updated_grads[0].max()
-        # mean loss terms
-        mean_grad = []
-        for ws in updated_grads:
-            ws_mean = ws.mean()
-            mean_grad += [1.0 if ws_mean == 0. else ws_mean]
-        return (ref_grad, mean_grad)
-
-    def update_loss_weights(self, epoch):
-        ref_grad, mean_grad = self.eval_loss_weights()
-        # evaluate new weights
-        # new weights
-        for i, wi in enumerate(self.model.loss_weights):
-            if i > 0:
-                # gp_weights[i]
-                gp_weight = ref_grad / (mean_grad[i] * K.get_value(wi))
-                new_val = (1.0-self.beta)*K.get_value(wi) + self.beta*gp_weight
-                K.set_value(self.model.loss_weights[i], new_val)
-        # print updates
-        print('adaptive_weights at epoch {}:'.format(epoch),
-              [K.get_value(wi) for wi in self.model.loss_weights])
-
