@@ -5,21 +5,166 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-import tensorflow.python.keras.backend as K
+from tensorflow.python.keras import backend as K
 graph_unique_name = K.get_graph().unique_name
 
-from tensorflow.python.keras.layers import Lambda
+from tensorflow.python.keras.layers import Lambda, Flatten, Dense
 from tensorflow.python.keras.layers import Dot
 from tensorflow.python.keras.layers import Input
 from tensorflow.python.keras.models import Model
+from tensorflow import sin as tf_sin
+from tensorflow import cos as tf_cos
+from tensorflow import tile as tf_tile
+from tensorflow import pow as tf_pow
+from tensorflow import divide as tf_divide
 
 from tensorflow import gradients as tf_gradients
 from tensorflow import stop_gradient as tf_stop_gradient
+from tensorflow import multiply, expand_dims
 
 from .utilities import *
 from .validations import *
 
 
+def fourier(f, w=10):
+    """Apply Fourier transform to the `Variable` objects..
+
+    # Arguments
+        w: (Int, np.ndarray) Frequencies of transformation.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+    append_to_bib("wang2020eigenvector")
+    layers = []
+    outputs = []
+    for fi in f.outputs:
+        if isinstance(w, int):
+            w = 2*np.pi/np.random.rand(w)
+        else:
+            assert isinstance(w, np.ndarray)
+        size_i = fi.shape.as_list()
+        assert len(size_i) == 2, 'Designed for the MLP model.'
+        assert size_i[-1] == 1, 'Only supported now.'
+        layers.append(
+            Dense(
+                w.size,
+                use_bias=False,
+                trainable=False,
+                activation=tf_sin,
+                name=graph_unique_name("fourier-sin")
+            )
+        )
+        outputs.append(layers[-1](fi))
+        layers[-1].set_weights(
+            [w.reshape(layers[-1].weights[0].shape.as_list())]
+        )
+        layers.append(
+            Dense(
+                w.size,
+                use_bias=False,
+                trainable=False,
+                activation=tf_cos,
+                name=graph_unique_name("fourier-cos")
+            )
+        )
+        outputs.append(layers[-1](fi))
+        layers[-1].set_weights(
+            [w.reshape(layers[-1].weights[0].shape.as_list())]
+        )
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(f.inputs.copy()),
+        outputs=outputs,
+        layers=layers
+    )
+
+    return res
+
+
+def monomial(f, p=10, w=1.):
+    """Apply monomial feature transformation to the `Variable` objects..
+
+    # Arguments
+        p: (Int, list) monomial powers to be considered.
+        w: (Int, list) weights for each term in the monomial(e.g. 1/n! for taylor expansion).
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+    if isinstance(p, int):
+        p = list(range(1, p+1))
+    else:
+        assert isinstance(p, (np.ndarray, list))
+    if isinstance(w, float):
+        w = len(p)*[w]
+    else:
+        assert isinstance(w, (np.ndarray, list))
+    layers = []
+    outputs = []
+    for fi in f.outputs:
+        f_dim = fi.shape.as_list()
+        tile_dim = (len(f_dim)-1)*[1] + [len(p)]
+        layers.append(
+            Lambda(lambda ys: tf_divide(tf_pow(tf_tile(ys, tile_dim), p), w),
+                   name=graph_unique_name("monomials"))
+        )
+        outputs.append(
+            layers[-1](fi)
+        )
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(f.inputs.copy()),
+        outputs=outputs,
+        layers=layers
+    )
+
+    return res
+
+
+def outer(a, b):
+    """outer product of two `Functional` objects.
+
+    # Arguments
+        a, b: outer(a,b)
+        Note that at least one of them should be of type Functional.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(a)
+    validate_functional(b)
+    layers = []
+    outputs = []
+    for a_out in a.outputs:
+        for b_out in b.outputs:
+            a_shape = a_out.shape.as_list()
+            b_shape = b_out.shape.as_list()
+            a_exp = len(a_shape)
+            b_exp = 1 if b_shape[0] is None else 0
+            name = graph_unique_name("outer")
+            layers.append(
+                Lambda(lambda ys: multiply(expand_dims(ys[0], a_exp),
+                                           expand_dims(ys[1], b_exp)), name=name)
+            )
+            net_output = layers[-1]([a_out, b_out])
+            layers.append(Flatten())
+            outputs.append(layers[-1](net_output))
+    # return the functional
+    assert a.get_class() == b.get_class()
+    Functional = a.get_class()
+    res = Functional(
+        inputs=unique_tensors(a.inputs.copy() + b.inputs.copy()),
+        outputs=outputs,
+        layers=layers
+    )
+    return res
+
+
 def pow(a, b):
     """Element-wise exponentiation applied to the `Functional` object.
 
@@ -49,7 +194,7 @@ def pow(a, b):
     res = Functional(
         inputs = unique_tensors(f.inputs.copy()),
         outputs = _apply_operation(lmbd, f),
-        layers = lmbd,
+        layers = lmbd
     )
 
     return res
@@ -683,6 +828,204 @@ def mean(x, **kwargs):
     return _apply_function(x, 'mean', **kwargs)
 
 
+def equal(f, other, tol=None):
+    """Element-wise comparison applied to the `Functional` objects.
+
+    # Arguments
+        f: Functional object.
+        other: A python number or a tensor or a functional object.
+        tol: (float) If you need a tolerance measure.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+    assert isinstance(tol, (type(None), float)), 'Expected a floating value for `tol`.'
+
+    inputs = f.inputs.copy()
+    if is_functional(other):
+        inputs += to_list(other.inputs)
+        if tol is None:
+            lambda_opr = lambda x: K.cast_to_floatx(K.equal(x[0], x[1]))
+        else:
+            lambda_opr = lambda x: K.cast_to_floatx(K.less_equal(K.abs(x[0]-x[1]), tol))
+    else:
+        _warn_for_ndarray(other)
+        if tol is None:
+            lambda_opr = lambda x: K.cast_to_floatx(K.equal(x, other))
+        else:
+            lambda_opr = lambda x: K.cast_to_floatx(K.less_equal(K.abs(x-other), tol))
+
+    lmbd = [Lambda(lambda_opr, name=graph_unique_name("equal")) for X in f.outputs]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(inputs),
+        outputs=_apply_operation(lmbd, f, other),
+        layers=lmbd
+    )
+    return res
+
+
+def not_equal(f, other, tol=None):
+    """Element-wise comparison applied to the `Functional` objects.
+
+    # Arguments
+        f: Functional object.
+        other: A python number or a tensor or a functional object.
+        tol: (float) If you need a tolerance measure.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+    assert isinstance(tol, (type(None), float)), 'Expected a floating value for `tol`.'
+
+    inputs = f.inputs.copy()
+    if is_functional(other):
+        inputs += to_list(other.inputs)
+        if tol is None:
+            lambda_opr = lambda x: K.cast_to_floatx(K.not_equal(x[0], x[1]))
+        else:
+            lambda_opr = lambda x: K.cast_to_floatx(K.greater(K.abs(x[0] - x[1]), tol))
+    else:
+        _warn_for_ndarray(other)
+        if tol is None:
+            lambda_opr = lambda x: K.cast_to_floatx(K.not_equal(x, other))
+        else:
+            lambda_opr = lambda x: K.cast_to_floatx(K.greater(K.abs(x - other), tol))
+
+    lmbd = [Lambda(lambda_opr, name=graph_unique_name("not_equal")) for X in f.outputs]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(inputs),
+        outputs=_apply_operation(lmbd, f, other),
+        layers=lmbd
+    )
+    return res
+
+
+def greater(f, other):
+    """Element-wise comparison applied to the `Functional` objects.
+
+    # Arguments
+        f: Functional object.
+        other: A python number or a tensor or a functional object.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+
+    inputs = f.inputs.copy()
+    if is_functional(other):
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater(x[0], x[1])), name=graph_unique_name("greater")) for X in f.outputs]
+    else:
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater(x, other)), name=graph_unique_name("greater")) for X in f.outputs]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(inputs),
+        outputs=_apply_operation(lmbd, f, other),
+        layers=lmbd
+    )
+    return res
+
+
+# legacy support
+tol_equal = equal
+
+
+def greater_equal(f, other):
+    """Element-wise comparison applied to the `Functional` objects.
+
+    # Arguments
+        f: Functional object.
+        other: A python number or a tensor or a functional object.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+
+    inputs = f.inputs.copy()
+    if is_functional(other):
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater_equal(x[0], x[1])), name=graph_unique_name("greater_equal")) for X in f.outputs]
+    else:
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater_equal(x, other)), name=graph_unique_name("greater_equal")) for X in f.outputs]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(inputs),
+        outputs=_apply_operation(lmbd, f, other),
+        layers=lmbd
+    )
+    return res
+
+
+def less(f, other):
+    """Element-wise comparison applied to the `Functional` objects.
+
+    # Arguments
+        f: Functional object.
+        other: A python number or a tensor or a functional object.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+
+    inputs = f.inputs.copy()
+    if is_functional(other):
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less(x[0], x[1])), name=graph_unique_name("less")) for X in f.outputs]
+    else:
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less(x, other)), name=graph_unique_name("less")) for X in f.outputs]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(inputs),
+        outputs=_apply_operation(lmbd, f, other),
+        layers=lmbd
+    )
+    return res
+
+
+def less_equal(f, other):
+    """Element-wise comparison applied to the `Functional` objects.
+
+    # Arguments
+        f: Functional object.
+        other: A python number or a tensor or a functional object.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+
+    inputs = f.inputs.copy()
+    if is_functional(other):
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less_equal(x[0], x[1])), name=graph_unique_name("less_equal")) for X in f.outputs]
+    else:
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less_equal(x, other)), name=graph_unique_name("less_equal")) for X in f.outputs]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs=unique_tensors(inputs),
+        outputs=_apply_operation(lmbd, f, other),
+        layers=lmbd
+    )
+    return res
+
+
 def _apply_function(x, fname, **kwargs):
     """Apply `fname` function to x element-wise.
 
@@ -1069,7 +1412,7 @@ def stop_gradient(f):
     res = Functional(
         inputs = unique_tensors(f.inputs.copy()),
         outputs = _apply_operation(lmbd, f),
-        layers = lmbd,
+        layers = lmbd
     )
 
     return res
